/* 
 * Graph Bisect Library
 * I/O Routines
 *
 * Grid Challenge 2006 Final
 * Team Leo - Nan DUN
 * Department of Computer Science, the University of Tokyo
 */

#include "gbimpl.h"

#define MAX_LINE			1280000
#define	MAX_VERTICES		10000
#define MAX_VERTICESdiv8	1250	/* Should be defined as MAX_VERTICES divided by 8 */
#define MAX_PREAMBLE		10000

unsigned long g_vertices, g_edges;
char g_bitmap[MAX_VERTICES][MAX_VERTICESdiv8];
char g_masks[8] = { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 };
static char g_preamble[MAX_PREAMBLE];

/* Conver METIS file to Johnson file */
void GB_Metis2Johnson(const char *filefrom, const char *fileto) {
	const char FUNCNAME[] = "GB_Metis2Johnson(2)";
	FILE *fromfp, *tofp;

	long node_cnt, edge_cnt, *adjLst, vwgt, ewgt;
	
	int fmt, readew, readvw, ncon, wgtflag;

	char line[MAX_LINE], *oldstr, *newstr;
	int setwidth, temp;
	long  i, j, k, l;
	
	fromfp = fopen(filefrom, "r");
	if (fromfp == NULL)	{
		GB_Error(GB_ERR_FILE, FUNCNAME);
	}
	
	tofp = fopen(fileto, "wb");
	if (tofp == NULL) {
		GB_Error(GB_ERR_FILE, FUNCNAME);
	}
	
	memset(line, 0x0, MAX_LINE);

	do 
	{
		fgets(line, MAX_LINE, fromfp);
	} while (line[0] == '%' && !feof(fromfp));

	if (feof(fromfp)) {
		if (fromfp != NULL)
			fclose(fromfp);
		if (tofp != NULL)
			fclose(tofp);
    	GB_Error(GB_ERR_FILE, FUNCNAME);
	}

	fmt = ncon = 0;
	sscanf(line, "%ld %ld %d %d", &node_cnt, &edge_cnt, &fmt, &ncon);

	readew = (fmt % 10 > 0);
	readvw = ((fmt / 10) % 10 > 0);
	if (fmt >= 100)	{
		GB_Error(GB_ERR_FILE, FUNCNAME);
	}


	wgtflag = 0;
	if (readew)
		wgtflag += 1;
	if (readvw)
		wgtflag += 2;


	if (ncon > 0 && !readvw) {
		GB_Error(GB_ERR_FILE, FUNCNAME);
	}

	adjLst = (long*) malloc(sizeof(long) * node_cnt);
	if (adjLst == NULL)	{
		GB_Error(GB_ERR_MEM, FUNCNAME);
	}
	
	edge_cnt = edge_cnt * 2;
	ncon = (ncon == 0 ? 1 : ncon);

	/* set width for output */
	setwidth = 1;
	temp = 10;
	for ( ; ; ) {
		if ((node_cnt / temp) > 0) {
			setwidth++;
			temp = temp * 10;
		} else {
			break;
		}
	}

	/* Start reading the graph file */
	for (k = 0, i = 0; i < node_cnt; i++) {
		do {
			fgets(line, MAX_LINE, fromfp);
		} while (line[0] == '%' && !feof(fromfp));
		oldstr = line;
		newstr = NULL;

		if (strlen(line) == MAX_LINE) {
			GB_Error(GB_ERR_OVER, FUNCNAME);
		}
		
		/* Output node ID and coordinates */
		fprintf(tofp, "%*ld (0.00000,0.00000)", setwidth, i + 1);
		
		if (readvw)	{
			for (l = 0; l < ncon; l++) {
				vwgt = strtol(oldstr, &newstr, 10);
				oldstr = newstr;
			}
		}
		
		j = 0;
		for (;;) {
			adjLst[j] = strtol(oldstr, &newstr, 10) - 1;
			oldstr = newstr;
			if (readew)	{
				ewgt = strtol(oldstr, &newstr, 10);
				oldstr = newstr;
			}
			if (adjLst[j] < 0)
				break;
			j++;
			k++;
		}

		/* Output adjacent nodes */
		fprintf(tofp, "%*ld ", setwidth, j);
		for (l = 0; l < j; l++)
			fprintf(tofp, "%ld ", adjLst[l] + 1);
		fprintf(tofp, "\n");
	}
	
	if (k != edge_cnt) {
		fprintf(stderr, "ERROR: Error in input file: %s.\n", filefrom);
		fprintf(stderr, "First line contained %ld edges, \
			but only found %ld edges in the file.\n", edge_cnt / 2, k / 2);
		GB_Error(GB_ERR_FILE, FUNCNAME);
	}

	if (adjLst != NULL)
		free(adjLst);
	if (fromfp != NULL)
		fclose(fromfp);
	if (tofp != NULL)
		fclose(tofp);

	return;
}

char getEdge(const int i, const int j) {
	int byte, bit;
	char mask;

	bit = 7 - (j & 0x00000007);
	byte = j >> 3;

	mask = g_masks[bit];
	return ((g_bitmap[i][byte] & mask) == mask);
}

int getParams(void) {
	const char FUNCNAME[] = "getParams()";
	char c, *tmp;
	char *pp = g_preamble;
	int stop;
	tmp = (char *)calloc(100, sizeof(char));
	if (tmp == NULL) {
		GB_Error(GB_ERR_MEM, FUNCNAME);
	}

	stop = 0;
	g_vertices = 0;
	g_edges = 0;

	while (!stop && (c = *pp++) != '\0') {
		switch (c) {
			case 'c':
				while ((c = *pp++) != '\n' && c != '\0');
				break;
				
			case 'p':
				sscanf(pp, "%s %lu %lu\n", tmp, &g_vertices, &g_edges);
				stop = 1;
				break;
				
			default:
				break;
		}
	}

	if (tmp != NULL)
		free(tmp);

	if (g_vertices == 0 || g_edges == 0) {
		fprintf(stderr, "Read parameters error.\n");
		return 0;
	} else {
		return 1;
	}
}

void GB_DimacsB2Johnson(const char *filefrom, const char *fileto) {
	const char FUNCNAME[] = "GB_Dimacs2Johnson(2)";
	FILE *fromfp, *tofp;
	unsigned long edge_cnt, i, j, temp;
	int length, setwidth;
	
	fromfp = fopen(filefrom, "r");
	if (fromfp == NULL)	{
		GB_Error(GB_ERR_FILE, FUNCNAME);
	}

	if (!fscanf(fromfp, "%d\n", &length)) {
		GB_Error(GB_ERR_FILE, FUNCNAME);
	}

	if (length >= MAX_PREAMBLE) {
		GB_Error(GB_ERR_FILE, FUNCNAME);
	}

	fread(g_preamble, 1, length, fromfp);
	g_preamble[length] = '\0';
	
	if (!getParams()) {
		GB_Error(GB_ERR_FILE, FUNCNAME);
	}

	/* Check g_vertices wheter exceed the MAX */
	if (g_vertices > MAX_VERTICES)	{
		GB_Error(GB_ERR_FILE, FUNCNAME);
	}
	for (i = 0; i < g_vertices; i++) {
		if (!fread(g_bitmap[i], 1, (int)((i + 8) / 8), fromfp))
			break;
	}
	if (fromfp != NULL)
		fclose(fromfp);
	
	/* Write to Johnson file */
	tofp = fopen(fileto, "wb");
	if (tofp == NULL) {
		GB_Error(GB_ERR_FILE, FUNCNAME);
	}	
	/* set width for output */
	setwidth = 1;
	temp = 10;
	for ( ; ; ) {
		if ((g_vertices / temp) > 0) {
			setwidth++;
			temp = temp * 10;
		} else {
			break;
		}
	}
	
	for (i = 0; i < g_vertices; i++) {
		edge_cnt = 0;
		for (j = 0; j < g_vertices; j++) {
			if (getEdge(i, j) || getEdge(j, i))
				edge_cnt++;
		}
		fprintf(tofp, "%*lu (0.000000,0.000000) %*lu ", setwidth, i + 1, setwidth, edge_cnt);
		for (j = 0; j < g_vertices; j++) {
			if (getEdge(i, j) || getEdge(j, i))
				fprintf(tofp, "%lu ", j + 1);
		}
		fprintf(tofp, "\n");
	}
	if (tofp != NULL)
		fclose(tofp);
	return;
}

/* Read graph file and fill graph structure */
void GB_ReadGraph(const char *filename, GB_GRAPH *graph) {
	const char FUNCNAME[] = "GB_ReadGraph(2)";

	FILE *fp;
	long offset;
	struct stat statBuf;
	char *fileBuf, *token;
	const char delimiters[] = ",() \t\n";
    
	char *file;
	integer *nv, *ne, ***adjlist, *vparity, *maxvdeg, *minvdeg, **belonglist, *edgecut; 
    integer vidx, vdeg, i, readflag;

	file = graph->filename;
    nv = &(graph->nv);
    ne = &(graph->ne);
    adjlist = &(graph->adjlist);
	vparity = &(graph->vparity);
    maxvdeg = &(graph->maxvdeg);
	minvdeg = &(graph->minvdeg);
    belonglist = &(graph->belonglist);
    edgecut = &(graph->edgecut);

    stat(filename, &statBuf);
    fp = fopen(filename, "r");
    if (fp == NULL) {
		GB_Error(GB_ERR_FILE, FUNCNAME);
    }
	memset(file, 0x0, GB_MAX_FILENAME);
	strcpy(file, filename);

    /* Greedy reading */
    fileBuf = (char *) malloc(sizeof(char) * (statBuf.st_size));
    if (fileBuf == NULL) {
    	GB_Error(GB_ERR_MEM, FUNCNAME);
	}
    fread(fileBuf, sizeof(char), statBuf.st_size, fp);

    /* Get the number of vertices from last line */
    /* Seek backward from EOF */
    offset = 0;
    for ( ; ; ) {
        /* find last '(' is more reliable instead of '\n'. */
        if (fileBuf[statBuf.st_size - offset] == '(')
            break;
        offset++;
    }
    /* Continue going back to last 2nd line's end. */
    for ( ; ; ) {
        if (fileBuf[statBuf.st_size - offset] == '\n') {
            sscanf(&(fileBuf[statBuf.st_size - offset]), "%*[^0-9+-]%d", nv);
            break;
        }
        offset++;
    }

	/* Allocate memory for all lists.
     * All lists start from 1. */
    *adjlist = (integer**) malloc(sizeof(integer*) * (*nv + 1));
    if (*adjlist == NULL) {
		GB_Error(GB_ERR_MEM, FUNCNAME);
    }

    *belonglist = (integer *) malloc(sizeof(integer) * (*nv + 1));
    if (*belonglist == NULL) {
    	GB_Error(GB_ERR_MEM, FUNCNAME);
	}

	readflag = *ne = *maxvdeg = 0;
	*minvdeg = *nv;
    for ( ; ; ) {
		if (readflag == 1) {
	        token = strtok(NULL, delimiters);
		} else {	/* first read should start from fileBuf */
			token = strtok(fileBuf, delimiters);
			readflag = 1;
		}
        if (token == NULL)
            break;

        vidx = atoi(token);
        if (vidx == 0) /* Double check EOF */
            break;

        (*belonglist)[vidx] = GB_PAR_L;

        /* coordinates no need */
        token = strtok(NULL, delimiters);
        token = strtok(NULL, delimiters);
		
		token = strtok(NULL, delimiters);
        vdeg = atoi(token);
		if (*maxvdeg < vdeg) {
            *maxvdeg = vdeg;
		}
		if (*minvdeg > vdeg) {
			*minvdeg = vdeg;
		}

        (*adjlist)[vidx] = (integer *) malloc(sizeof(integer) * (vdeg + 1));
        if ((*adjlist)[vidx] == NULL) {
        	GB_Error(GB_ERR_MEM, FUNCNAME);
		}
        *ne += vdeg;
        (*adjlist)[vidx][0] = vdeg;
        for (i = 1; i <= vdeg; i++) {
            token = strtok(NULL, delimiters);
            (*adjlist)[vidx][i] = (integer) atoi(token);
        }
    }
	if (*nv % 2 == 0) {
		*vparity = GB_NUM_EVEN;
	} else {
		*vparity = GB_NUM_ODD;
	}
    *edgecut = *ne = (*ne) / 2;

    if (fp != NULL)
        fclose(fp);
}

void GB_FreeGraph(GB_GRAPH *graph) {
	const char FUNCNAME[] = "GB_FreeGraph(1)";
	
    integer *nv, *ne, ***adjlist, **belonglist, *edgecut;
    integer i;

    nv = &(graph->nv);
    ne = &(graph->ne);
    adjlist = &(graph->adjlist);
    belonglist = &(graph->belonglist);
    edgecut = &(graph->edgecut);
	
	if (adjlist == NULL || belonglist == NULL) {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}
    for (i = 1; i <= *nv; i++) {
        if ((*adjlist)[i] != NULL) {
            free((*adjlist)[i]);
            (*adjlist)[i] = NULL;
        }
    }
    if (*adjlist != NULL) {
        free(*adjlist);
        *adjlist = NULL;
    }
    if (*belonglist != NULL) {
        free(*belonglist);
        *belonglist = NULL;
    }
}

/* Print graph summary or full structure */
void GB_PrintGraph(GB_GRAPH *graph) {
	const char FUNCNAME[] = "PrintGraph(3)";
	char *file;
	integer *nv, *ne, ***adjlist, *vparity, *maxvdeg, *minvdeg, **belonglist, *edgecut; 
	integer leftcount, rightcount, edgecutcount, i, j;

	if (graph == NULL) {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}
	
	file = graph->filename;
    nv = &(graph->nv);
    ne = &(graph->ne);
    adjlist = &(graph->adjlist);
	vparity = &(graph->vparity);
    maxvdeg = &(graph->maxvdeg);
	minvdeg = &(graph->minvdeg);
    belonglist = &(graph->belonglist);
    edgecut = &(graph->edgecut);

	leftcount = rightcount = edgecutcount = 0;
    for (i = 1; i <= *nv; i++) {
        /* Count bisection */
        if ((*belonglist)[i] == GB_PAR_L) {
            leftcount++;
            for (j = 1; j <= (*adjlist)[i][0]; j++) {
                if ((*belonglist)[((*adjlist)[i][j])] == GB_PAR_R)
                    edgecutcount++;
            }
        } else {
			rightcount++;
        }
    }
	fprintf(stdout, "------------------------------------------------------------------\n");
	if (file != NULL) {
		fprintf(stdout, "\tGraph: %s\n", file);
	}
	if (*vparity == GB_NUM_EVEN) {
		fprintf(stdout, "\tVertices [EVEN]: %d, Edges: %d\n", *nv, *ne);
	} else {
		fprintf(stdout, "\tVertices [ODD]: %d, Edges: %d\n", *nv, *ne);
	}
	
	fprintf(stdout, "\tVertex degree: Max: %d, Min: %d, Avg: %.4f\n", *maxvdeg, *minvdeg, ((double)(*ne)) / (*nv));
	fprintf(stdout, "\tI/O time: %f seconds\n", ((double)(g_io_end - g_io_start)) / CLOCKS_PER_SEC);

	if ((leftcount == *nv && rightcount == 0) || (leftcount == 0 && rightcount == *nv)) {
		fprintf(stdout, "\tGraph has not been bisected!\n");
		fprintf(stdout, "------------------------------------------------------------------\n");
		return;
	}

	fprintf(stdout, "\n\t");
	switch (g_algo) {
		case GB_GG:
			fprintf(stdout, "Graph bisected by Greedy Growing (GG) algorithm.\n");
			break;
		case GB_KL:
			fprintf(stdout, "Graph bisected by Kernighan-Lin (KL) algorithm.\n");
			break;
		case GB_LKL:
			fprintf(stdout, "Graph bisected by Linear Kernighan-Lin (LKL) algorithm.\n");
			break;
		case GB_MMG:
			fprintf(stdout, "Graph bisected by Min-Max Greedy (MMG) algorithm.\n");
			break;
		case GB_RTS:
			fprintf(stdout, "Graph bisected by Randomized Tabu Search (RTS) algorithm.\n");
			break;
		case GB_RRTS:
			fprintf(stdout, "Graph bisected by Reactive Randomized Tabu Search (RRTS) algorithm.\n");
			break;
		default:
			break;
	}
	fprintf(stdout, "\tBisection: Left: %d, Right: %d\n", leftcount, rightcount);
	if ((*vparity == GB_NUM_EVEN) && (leftcount != rightcount)) {
		fprintf(stdout, "\tEven number of vertices, but two partition are not equal.\n");
	}
	if ((*vparity == GB_NUM_ODD) && !((leftcount - rightcount) == 1 || (rightcount - leftcount) == 1)) {
		fprintf(stdout, "\tODD number of vertices, but difference does not equal one.\n");
	}

    fprintf(stdout, "\tEdge cut: %d\n", edgecutcount);

	if (edgecutcount != *edgecut) {
        fprintf(stdout, "\tNOT Match: I counted %d edge-cut, but %d in graph.\n", edgecutcount, *edgecut);
	}
	fprintf(stdout, "\tBisect time: %f seconds\n", ((double)(g_bi_end - g_bi_start)) / CLOCKS_PER_SEC);

    fprintf(stdout, "------------------------------------------------------------------\n");
}

