/* 
 * Graph Bisect Library
 * Utility Routines
 *
 * Grid Challenge 2006 Final
 * Team Leo - Nan DUN
 * Department of Computer Science, the University of Tokyo
 */

#include "gbimpl.h"

void GB_Error(int errno, const char *FUNCNAME) {
	char errstring[128];

	memset(errstring, 0x0, 128);
	switch (errno) {
		case GB_ERR_MEM:
			sprintf(errstring, "ERROR: Memory Allocation in %s.\n", FUNCNAME);
			break;

		case GB_ERR_OVER:
			sprintf(errstring, "ERROR: Overflow in %s.\n", FUNCNAME);
			break;

		case GB_ERR_NULL:
			sprintf(errstring, "ERROR: NULL pointer in %s.\n", FUNCNAME);
			break;

		case GB_ERR_FILE:
			sprintf(errstring, "ERROR: FILE I/O in %s.\n", FUNCNAME);
			break;

		case GB_ERR_ARG:
			sprintf(errstring, "ERROR: Wrong argument in %s.\n", FUNCNAME);
			break;

		default:
			sprintf(errstring, "UNKNOWN ERROR in %s.\n", FUNCNAME);
			break;
	}
	fprintf(stderr, errstring);
	exit(errno);
}

integer GB_Random(integer start, integer step) {
	const char FUNCNAME[] = "GB_Random(2)";

	if (step > RAND_MAX) {
		GB_Error(GB_ERR_OVER, FUNCNAME);
	}

	return ((integer)(rand() % step + start));
}

void GB_RandomPermute(integer *list, integer n) {
	const char FUNCNAME[] = "GB_RandomPermute(2)";
	integer pos, i;

	if (list != NULL) {
		for (i = 1; i <= n; i++) {
			list[i] = -1;
		}
		for (i = 1; i <= n; i++) {
			do {
				pos = GB_Random(1, n);
			} while (list[pos] != -1);
			list[pos] = i;
		}
	} else {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}
}

void GB_RandomPermuteExclude(integer *list, integer n, integer ex1, integer ex2) {
	const char FUNCNAME[] = "GB_RandomPermuteExclude(4)";
	integer pos, i;

	if (list != NULL) {
		for (i = 1; i <= n; i++) {
			list[i] = -1;
		}
		
		if (ex1 == ex2) {
			list[1] = ex1;
		} else {
			list[1] = ex1;
			list[2] = ex2;
		}
		for (i = 1; i <= n; i++) {
			if (i == ex1 || i == ex2) {
				continue;
			}
			do {
				pos = GB_Random(1, n);
			} while (list[pos] != -1);
			list[pos] = i;
		}
	} else {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}
}

void GB_BucketInitial(bucket_t **buckets, integer nb, itemp_t **items, integer ni, integer value) {
	const char FUNCNAME[] = "GB_BucketInitial(5)";
	integer i;
	
	*buckets = (bucket_t *) malloc(nb * sizeof(bucket_t));
	if (*buckets == NULL) {
		GB_Error(GB_ERR_MEM, FUNCNAME); 
	}
	
	*items = (itemp_t *) malloc(ni * sizeof(itemp_t));
	if (*items == NULL) {
		GB_Error(GB_ERR_MEM, FUNCNAME);
	}

	for (i = 0; i < nb; i++) {
		(*buckets)[i].head = (*buckets)[i].tail = NULL;
	}

	for (i = 0; i < ni; i++) {
		(*items)[i].value = value;
		(*items)[i].ptr = (item_t *) malloc(sizeof(item_t));
		if ((*items)[i].ptr == NULL) {
			GB_Error(GB_ERR_MEM, FUNCNAME);
		}
		(*items)[i].ptr->index = i;
		(*items)[i].ptr->prev = (*items)[i].ptr->next = NULL;
	}
}

void GB_BucketFinalize(bucket_t **buckets, itemp_t **items, integer ni) {
	const char FUNCNAME[] = "GB_BucketFinalize(4)";
	integer i;

	if (*buckets == NULL || *items == NULL) {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}

	for (i = 0; i < ni; i++) {
		if ((*items)[i].ptr != NULL) {
			free((*items)[i].ptr);
		}
	}
	free(*items);
	free(*buckets);
	*items = NULL;
	*buckets = NULL;
}

void GB_BucketReset(bucket_t *buckets, integer nb, itemp_t *items, integer ni, integer value) {
	const char FUNCNAME[] = "GB_BucketFinalize(5)";
	integer i;
	
	if (buckets == NULL || items == NULL) {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}

	for (i = 0; i < nb; i++) {
		buckets[i].head = buckets[i].tail = NULL;
	}
	for (i = 0; i < ni; i++) {
		items[i].value = value;
		items[i].ptr->prev = items[i].ptr->next = NULL;
	}
}

void GB_BucketInsert(integer value, integer index, bucket_t *buckets, integer *min, integer *max, itemp_t *items) {
	const char FUNCNAME[] = "GB_BucketInsert(6)";
	item_t **head, **tail, *curr;

	head = &(buckets[value].head);
	tail = &(buckets[value].tail);
	curr = items[index].ptr;
	items[index].value = value;

	if (curr == NULL) {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}

	if (*head == NULL) {	/* buckets is empty */
		*head = *tail = curr;
		curr->prev = curr->next = NULL;
	} else {					/* buckets is not empty, append to tail */
		(*tail)->next = curr;
		curr->prev = *tail;
		curr->next = NULL;
		*tail = (*tail)->next;
	}

	/* setup min and max */
	if (value < *min) {
		*min = value;
	}
	if (value > *max) {
		*max = value;
	}
}

void GB_BucketDelete(integer value, integer index, bucket_t *buckets, integer *min, integer *max, itemp_t *items) {
	const char FUNCNAME[] = "GB_BucketDelete(6)";
	item_t **head, **tail, *curr;
	
	head = &(buckets[value].head);
	tail = &(buckets[value].tail);
	curr = items[index].ptr;
	items[index].value = -1;
	
	if (curr == NULL) {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}
	
	if (*head == *tail) {		/* last node in bucket[value] */
		*head = *tail = NULL;
		curr->prev = curr->next = NULL;
		if (value == *min) {
			if (*min == *max) {	/* the very last node in buckest */
				*min = INT_MAX;
				*max = 0 - INT_MAX;
				return;
			}
			while (*min <= *max) {
				*min += 1;
				if (buckets[*min].head != NULL)
					break;
			}
		}

		if (value == *max) {	/* *max cannot be equal to *min, we have handled it above */
			while (*min <= *max) {
				*max -= 1;
				if (buckets[*max].head != NULL) {
					break;
				}
			}
		}
		return;
	}

	if (*head == curr) {	/* delete from head */
		*head = curr->next;
		(curr->next)->prev = NULL;
		curr->prev = curr->next = NULL;
		return;
	}

	if (*tail == curr) {	/* delete from tail */
		*tail = (*tail)->prev;
		(*tail)->next = curr->prev = NULL;
		return;
	}

	/* delete from middle */
	(curr->prev)->next = curr->next;
	(curr->next)->prev = curr->prev;
	curr->prev = curr->next = NULL;
}

integer GB_BucketGetFirstMin(bucket_t *buckets, integer min) {
	const char FUNCNAME[] = "GB_BucketGetFirstMin(2)";
	item_t *curr;

	curr = buckets[min].head;
	if (curr == NULL) {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}
	
	return curr->index;
}

integer GB_BucketGetFirstMax(bucket_t *buckets, integer max) {
	const char FUNCNAME[] = "GB_BucketGetFirstMax(2)";
	item_t *curr;
	
	curr = buckets[max].head;
	if (curr == NULL) {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}

	return curr->index;
}

void GB_SeedDistribute(int hostnum, int hostid, integer nv, int *seed) {
	int step, remains;
	
	seed[0] = 1;
	seed[1] = nv;
	if (hostnum == 1) {
		return;
	}

	if (hostnum >= nv) {
		seed[0] = seed[1] = (hostid - 1) % nv + 1;
	} else {
		step = nv / hostnum;
		remains = nv - (step * hostnum);
		if (hostid <= remains){
			seed[0] = (hostid - 1) * (step + 1) + 1;
			seed[1] = hostid * (step + 1);
		} else {
			seed[0] = remains + (hostid - 1)* step + 1;
			seed[1] = remains + hostid * step;
		}
	}

}

